; ----------------------------------------------------------------
; The Rx Core
; ----------------------------------------------------------------
display "Rx :: Release 0",0Dh,0Ah	; Assembly-time; display the version
include "rx.macros"			; Include the standard Rx macros


bootstrap_rx:				; Initialize Rx
	call reset			; This sets up the data stack
	upsh bootstrap			; and evaluates the bootstrap
	upsh [bootstrap.size]		; code.
	call eval			;
next


self '(', comment			;
	upsh ')'			; Parse the input stream until  
	call parse			; the ) symbol is found; ignore
	drop				; everything between ( and )
	drop				;
next

; --------------------------------------------------------------
; This group of words implement the primary word classes and
; wrapper functions to help with readability.
; --------------------------------------------------------------

code 'forth', dict_forth		; These routines are used to
	mov dword [class], do_forth	; set the active class for
next					; new definitions.
code 'macro', dict_macro		; 
	mov dword [class], do_macro	; You can also do this with
next					; ' class-name class !
code 'self', dict_self			; but this is a bit more
	mov dword [class], do_self	; readable.
next					;
code 'inline', dict_inline		;
	mov dword [class], do_inline	;
next					;


ccode '.class', do_class		; .class class
	jmp do_forth			; Same action as the .forth class


ccode '.data', do_data			; .data class
	cmp dword [state], -1		;
	jz literal			;
next					;


ccode '.forth', do_forth		; .forth class
	cmp dword [state], -1		;
	jz compile			;
					; falls through into .self

ccode '.self', do_self			; .self class
	upop edx			;
	jmp edx				;


ccode '.macro', do_macro		; .macro class
	cmp dword [state], -1		;
	jz do_self			;
	drop				;
next					;


ccode '.inline', do_inline		; .inline class
	cmp dword [state], -1		;
	jnz do_self			; Run the word at interpret time
					;
.get:	mov ebx, [eax]			; At compile time, copy the compiled
	and ebx, 0xff			; definition into the current
.check:	cmp ebx, 0xc3			; definition.
	jz .done			;
.place:	upsh ebx			; This will not work with words
	call comma1			; containing CALLs to other words
.next:	inc eax				; or the $c3 byte in their definition
	jmp .get			;
					;
.done:	drop				; When done, clean up and exit
next					;
; --------------------------------------------------------------


; --------------------------------------------------------------
; This group of words deals with the creation, location, and
; manipulation of dictionary headers.
; --------------------------------------------------------------


; -------------------------------------	;
; Dictionary entries take the following	;
; form:					;
;					;
; offset    description          word   ;
; ------------------------------------- ;
;  0     link to previous entry	.link	;
;  4     xt of word		.xt	;
;  8     class of word		.class	;
; 12     length of word name	.name	;
; 13+    word name			;
; ------------------------------------- ;


code 'entry', _entry			; Create a new dictionary entry
	push ecx	        	;
	mov edi,[d]	        	; LFA
	mov ecx,[last]	        	; last
	mov edx,[ecx]	       		;
	mov [edi],edx	        	; LFA= [last]
	mov [ecx],edi	        	; last= LFA
	mov ecx,[h]	       		;
	mov [edi+4],ecx         	; CFA= here
	mov edx, [class]		;
	mov [edi+8],edx			; CLASS
	mov [edi+12],al	        	; Length
	add edi,13	        	;
	upop ecx	        	;
	push esi	        	;
	mov esi,eax	        	;
	rep movsb	        	;
	mov [d],edi	        	; d= d+13+length
	pop esi 	        	;
	pop ecx 	        	;
	drop				;
next					;


code 'find', find	        	; Find a word in the dictionary
	push ebx                	;
	mov ebx, flast          	;
.o:	upop ecx                	;
	push esi			;
	mov esi,TOS             	;
					;
	mov dh, byte [TOS]     		; We get the first char and the
	mov dl, cl              	; length, and compare both at
.a:	mov ebx,[ebx]           	; once. This ensures that we only
	test ebx,ebx            	; check entries that might be right.
	jz .end                 	; end of wordlist
	lea edi, [ebx+13]       	; point edi at the string
	cmp  dx, [edi-1]		;
	jne .a				;
.len:	push esi                	; same length, so do the compare
	push ecx               		;
	repe cmpsb              	;
	pop ecx                 	;
	pop esi                 	;
	jne .a                  	; If they weren't identical, try
					; again.
	mov dword [which], ebx		; Otherwise set "which" to the entry
        mov eax,[ebx+8]         	;
        mov dword [action], eax 	; "action" is set to this word's class
	mov eax,[ebx+4]         	; Obtain the XT
.ret:   pop esi				;
	pop ebx                 	;
	upsh -1				; Return TRUE
	ret				;
.end:	pop esi				;
	upsh ecx                	; No matches were found
	upsh 0				; Return FALSE
	pop ebx				;
next
; --------------------------------------------------------------


; --------------------------------------------------------------
; Dealing with numbers, interpret-time
; --------------------------------------------------------------

code '>number', number	        	; >number
					; This routine attempts to convert
					; a counted string into a number.
	push dword [base]		;
	mov ecx,TOS	        	; n  (keep on stack in case of failure)
	mov ebx,NOS	        	; a
	dup		        	;
	xor eax,eax	        	; the number
	xor edx,edx	        	; temp
	mov dl,[ebx]	        	;
	cmp dl,45	        	; -     Sign prefix
	pushf		        	;
	jne .a		        	;
	inc ebx 	        	;
	dec ecx 	        	;
	mov dl,[ebx]	        	;
.a:	sub dl,35	        	; #$%&'  Base prefix
	cmp dl,5	        	;
	ja .b		        	;
	mov dl,[edx+bases]      	;
	mov byte [base],dl      	;
	inc ebx 	        	;
	dec ecx 	        	;
.b:	mov dl,[ebx]	        	; digits
	inc ebx 	        	;
	call digit	        	;
	jc .err 	        	;
	loop .b 	        	;
	jmp .c		        	;
.err:	add esp, 4	        	;
	drop		        	;
	upsh 0				; FALSE, not a number
	jmp .ret	        	;
.c:	popf		        	;
	jne .d		        	;
	neg eax 	        	;
.d:	add esi,8	        	;
	upsh -1				; TRUE, was a number
.ret:	pop dword [base]        	;
next


digit:	cmp byte [base],255     	;
	je .x10 	        	;
	cmp dl,57	        	; 9
	jbe .a		        	;
	and dl,5Fh	        	; uppercase
	cmp dl,65	        	; throw out chars between '9' and 'A'
	jb .err 	        	;
	sub dl,7	        	;
.a:	sub dl,48	        	; 0
	cmp dl, byte [base]     	;
	jb .x10 	        	;
.err:	stc		        	; not a digit
	ret		        	;
.x10:	imul eax,[base]         	;
	add eax,edx	        	;
	clc		        	;
next
; --------------------------------------------------------------



; --------------------------------------------------------------
; The compiler 
; --------------------------------------------------------------

self ':', colon 	        	; Ok, this is the entry to the compiler
	mov ebx, dword [h]		;
	mov dword [old_h], ebx		;
	mov dword [tail], -1		; disable tail recursion
	upsh [whitespace]		;
	call parse			;	
	call _entry	        	; Create a new word
					; Fall through into ]
					;
code ']', rbracket			; Switch to compile mode
	mov dword [state], -1		; 
next					;


mcode ';', semi 			; Compile in an exit
	call ssemi	        	; And go back to the interpreter
					; Fall through into [
					;
mcode '[', lbracket	        	; Switch to interpret mode
	mov dword [state], 0		;
next					;


mcode ';;', ssemi	        	; Exit a word (; will call this)
	cmp dword [tail], 1		;
	jz .b				;
	mov edx,[h]	        	;
	sub edx,5	        	;
	cmp byte [edx],0xE8     	; Was the last thing compiled a CALL?
	jnz .a		        	; No, skip the next line
	cmp dword [tail],0		; See if we should compile a tail-call
	jnz .a				; No? skip it and compile a ret
	inc byte [edx]	        	; Yes, change to JMP
	mov dword [tail], 1		;
	next		        	; and exit
.a:	mov byte [edx+5],0xC3   	; If not a CALL, compile in a RET
	inc dword [h]	        	;
	mov dword [tail], -1		;
.b:	next


code 'compile', compile         	; This routine compiles in CALL's to
	sub TOS,[h]	        	; words. Address is passed in EAX
	sub TOS,5	        	;
	upsh 0xE8	        	; CALL opcode
	call comma1	        	; 
	call comma	        	; ADDRESS
	mov dword [tail], 0		; Enable tail-call elimination
next					;



; -------------------------------------	;
; Literals take the following form by   ;
; default:                              ;
;                                       ;
; call dolit                            ;
; dd   value                            ;
;                                       ;
; The dolit subroutine handles skipping ;
; over the value.                       ;
; -------------------------------------	;

mcode 'literal', literal        	; Compile in a literal. See
	upsh dolit	        	; the diagram above for how
	call compile	        	; this looks in memory.
	call comma	        	;
next					;


dolit:	dup		        	; This is where we handle literals.
	mov TOS,[esp]	        	; Two steps:
	mov TOS,[TOS]	        	;  1) Fetch the value from the return
					;     stack
	add dword [esp],4       	;  2) Skip to the next instruction
next					;     



code ',', comma 	        	; comma (,) saves a value to "here"
	mov ecx,4	        	; By default, it uses a dword (4 bytes)
.a:	mov edx,[h]	        	;
	mov [edx],eax	        	;
	drop		        	;
	add edx,ecx	        	;
	mov [h],edx	        	;
	mov dword [tail], -1		; Disable tail recursion
next					;


code '1,', comma1	        	; comma1 (1,) saves 1 byte to "here"
	mov ecx,1	        	;
	jmp comma.a			;


code '2,', comma2	        	; comma2 (2,) saves 2 bytes to "here"
	mov ecx,2	        	;
	jmp comma.a			;


code '3,', comma3	        	; comma3 (3,) saves 3 bytes to "here"
	mov ecx,3	        	;
	jmp comma.a			;
; --------------------------------------------------------------


; --------------------------------------------------------------
; The interpreter (actually, this is is also used during
; compilation. The classes determine how things are handled
; based on the current compiler state).
; --------------------------------------------------------------

interpret:
.word:	upsh [whitespace]		; Look for whitespace
	call parse			; Get a word from the TIB
	jnz .find	        	; Look for the end of the line.
	drop		        	; If at EOL, clean up the stack
	drop		        	; and exit
	next				;

.find:	call find	        	; See if we can find the word

.exec:	cmp TOS, 0			;
	drop				; Get rid of the flag
	jz .lit				; 
	mov edx, [action]		; If the word was found, call the 
	call edx			; class handler for it.
	jmp .word	        	; 

.lit:	call number	        	; No? is it a number?
	cmp TOS, 0			;
	drop				; Get rid of the flag
	jz .error      			; 
	call do_data			; Numbers are a .data type
	jmp .word

.error:	call notfound			; Couldn't find it -- Call "word?"
	cmp TOS, 0			;
	drop				;
	jnz .word			;
next					; Exit if "word?" couldn't handle it


query:					;
	mov dword ecx,[source]  	; Look for the end of the input stream
	or ecx,ecx	        	;
	jz .exit	        	;
	mov edi,[tin]	        	;
	upsh edi	        	; Input pointer
	dup		        	;
	cmp byte [edi], 9		; TAB
	jne .notab			;
	mov byte [edi], 32		; Make the TAB a SPACE
.notab:	cmp byte [edi], 10       	; Skip LF
	jne .a				;
	inc edi 	        	; Increase our pointer
.a:	mov [tin],edi	        	;
	sub ecx,edi	        	; Remaining length
	jbe .eof	        	;
	mov al,10	        	; Line Feed
	repne scasb	        	;
	mov eax,edi	        	;
	jne .b		        	;
	dec eax 	        	;
	cmp Byte [eax-1],13     	; We make CR optional
	jne .b		        	;
	dec eax 	        	;
.b:	upop [tp]	        	;
	drop		        	;
	next		        	;
.eof:	drop		        	;
	drop		        	;
	add esp,4	        	; Discard caller
	pop Dword [tp]	        	;
	pop Dword [tin]         	;
	pop Dword [source]      	;
.exit:					;
next					;


code 'eval', eval			;
	push Dword [source]     	; Save "source"
	push Dword [tin]        	; Save ">in"
	push Dword [tp]         	; Save "tp"
	add  TOS,NOS	        	;
	upop [source]	        	; New "source"
	upop [tin]	        	; New ">in"
.a:	call query			;
	call interpret			;
	jmp .a				;


code 'parse', parse	        	;
	mov edi,[tin]	        	; Pointer into TIB
	mov ecx,[tp]	        	;
	sub ecx,edi	        	;
	inc ecx 	        	;
	repe scasb	        	;
	dec edi 	        	;
	inc ecx 	        	;
	dup		        	;
	mov NOS,edi	        	; address
	repne scasb	        	;
	mov TOS,edi	        	;
	dec TOS 	        	;
	sub TOS,NOS	        	; length
	mov [tin],edi	        	;
next					;
; --------------------------------------------------------------


; --------------------------------------------------------------
; Other words that don't fit neatly into the above categories
; --------------------------------------------------------------

code 'reset', reset	        	; This word is used to reset the stack
	mov ecx, 2048			;
	mov esi, s0			;
	sub esi, 1024			;
.empty:	mov byte [esi], 0		;
	inc esi				;
	loop .empty			;
	mov esi, s0			;
	xor TOS, TOS			;
next					;


code 'last', var_last			;
	upsh [last]			;
next					;


code 'tib', _tib			;
	upsh tib			;
next					;


code 'word?', notfound			;
	drop				;
	drop				;
	upsh 0				; FALSE, unable to handle the error
next					;


; --------------------------------------------------------------
var 'h0', h, h0		        	; h0     (pointer to HERE)
var 'base', base, 10	        	; base   current numeric base
var '>in', tin, 0	        	; >IN    Pointer into the TIB
var 'class', class, 0			;
var 'state', state, 0			;
var 'which', which, 0			;
var 'whitespace', whitespace, 32	; What we look for as a separator

action		dd 0			; Handler for the current class
old_h		dd 0			;
d		dd d0			;
tail		dd 0			; Allow tail-calls?
source		dd 0	        	; Evaluate from RAM or KBD
tp		dd tib	        	; TP (pointer to input buffer)
bases		db 10,16,2,8,255	; #dec $hex %bin &oct 'ascii
; --------------------------------------------------------------
align 4					;
bootstrap:				;
	file "rx.forth"			;
.size dd $-bootstrap			;
; --------------------------------------------------------------
include 'rx.dict'
align 4
last  dd flast				; Last word in dictionary
flast dd forth_link	        	; Last word in 'forth'
; --------------------------------------------------------------
include "rx.cfg"			;
; --------------------------------------------------------------
